/* adv-driver.c
 * 
 * Összetett eszközvezérlő
 * 
 * Az 1. állomány egy valódi állományt szimulál, amelynek mérete 256 karakter.
 *
 * A 2. állomány szintén egy valodi állományt szimulál, azonban kezeli a
 * hozzáírást és a felülírást is. Továbbá támogatja ioctl függvénnyel az állomány
 * méretének lekérdezését és módosítását, valamint a seek és az mmap függvényeket.
 *
 * A 3. állomány hasonlít a 2-hoz azonban az olvasást blokkolja, amíg nem irják be
 * az igényelt mennyiségű adatot. Továbbá az ioctl mellett a select-et is támogatja.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/mm.h>
#include <linux/gfp.h>
#include <linux/io.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/device.h>

#include "adv-driver.h"

#define DEVICE_NAME "hello"
static int major_num=120;
#define CLASS_NAME "helloclass"
static struct class* hello_class;

#define NUM_OF_DEVS 3

typedef struct HelloDev 
{
  //char* data;
  unsigned long data;
  unsigned long size;
  struct semaphore lock;
} HelloDev;

HelloDev helloDevs[NUM_OF_DEVS];

wait_queue_head_t hello_queue;

int hello_open(struct inode *inode, struct file *pfile);
int hello_close(struct inode *inode, struct file *pfile);
ssize_t hello1_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos);
ssize_t hello1_read(struct file *pfile, char *buf, size_t count, loff_t *ppos);
ssize_t hello2_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos);
ssize_t hello2_read(struct file *pfile, char *buf, size_t count, loff_t *ppos);
static int hello_mmap(struct file* pfile, struct vm_area_struct *vma);
ssize_t hello3_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos);
ssize_t hello3_read(struct file *pfile, char *buf, size_t count, loff_t *ppos);
unsigned int hello3_poll(struct file *pfile, poll_table *wait);
long hello_ioctl(struct file *pfile, unsigned int cmd, unsigned long arg);
loff_t hello_llseek(struct file *pfile, loff_t off, int whence);

struct file_operations hello1_fops =
{
  owner:	THIS_MODULE,
  read:	hello1_read,
  write:	hello1_write,
  open:	hello_open,
  release:	hello_close
};

struct file_operations hello2_fops =
{
  owner:	THIS_MODULE,
  read:	hello2_read,
  write:	hello2_write,
  unlocked_ioctl:	hello_ioctl,
  llseek:	hello_llseek,
  open:	hello_open,
  release:	hello_close,
  mmap: hello_mmap
};

struct file_operations hello3_fops =
{
  owner:	THIS_MODULE,
  read:	hello3_read,
  write:	hello3_write,
  unlocked_ioctl:	hello_ioctl,
  poll:	hello3_poll,
  open:	hello_open,
  release:	hello_close
};

struct file_operations *hello_fops_array[] =
{
  &hello1_fops,
  &hello2_fops,
  &hello3_fops
};

int hello_open(struct inode *inode, struct file *pfile)
{
  int num;
  
  num=MINOR(inode->i_rdev);
  if(num>=NUM_OF_DEVS) return -ENODEV;
  
  if(pfile->private_data==NULL)
  {
    pfile->f_op=hello_fops_array[num];
    pfile->private_data=&helloDevs[num];
  }
  
  try_module_get(THIS_MODULE);
  
  printk(KERN_INFO DEVICE_NAME " open.\n");

  if(num == 1)
  {
    if(pfile->f_flags & O_TRUNC)
    {
      helloDevs[num].size=0;
    }
  }
  
  return 0;
}

int hello_close(struct inode *inode, struct file *pfile)
{
  printk(KERN_INFO DEVICE_NAME " close.\n");
  module_put(THIS_MODULE);
  return 0;
}

ssize_t hello1_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos)
{
  HelloDev* pdev=pfile->private_data;
  ssize_t res = -ENOMEM;
  
  if(count == 0) return 0;
  if(*ppos >= BUFFER_SIZE) return -ENOMEM;
  
  if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;

  if(pdev->data == 0) goto error;
  
  if((*ppos+count) > BUFFER_SIZE) count=BUFFER_SIZE-*ppos;
    
  if(copy_from_user((char*)(pdev->data)+*ppos, buf, count))
  {
    res=-EFAULT;
    goto error;
  }
  *ppos+=count;
  res=count;

  if(pdev->size < *ppos) pdev->size=*ppos;
  
error:  
  up(&pdev->lock);
  
  return res;
}

ssize_t hello1_read(struct file *pfile, char *buf, size_t count, loff_t *ppos)
{
  HelloDev* pdev=pfile->private_data;
  ssize_t res = 0;
  
  if(count == 0) return 0;
  
  if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;

  if(pdev->data == 0) goto error;
  if(*ppos >= pdev->size) goto error;
  if((*ppos+count) > pdev->size) count=pdev->size-*ppos;
    
  if(copy_to_user(buf, (char*)(pdev->data)+*ppos, count))
  {
    res=-EFAULT;
    goto error;
  }
  *ppos+=count;
  res=count;
  
error:  
  up(&pdev->lock);
  
  return res;
}

ssize_t hello2_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos)
{
  HelloDev* pdev=pfile->private_data;
  ssize_t res = -ENOMEM;
  
  if(count == 0) return 0;
  if(*ppos >= BUFFER_SIZE) return -ENOMEM;
  
  if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;

  if(pdev->data == 0) goto error;
  
  if(*ppos == 0) *ppos=pdev->size;
  if((*ppos+count) > BUFFER_SIZE) count=BUFFER_SIZE-*ppos;
    
  if(copy_from_user((char*)(pdev->data)+*ppos, buf, count))
  {
    res=-EFAULT;
    goto error;
  }
  *ppos+=count;
  res=count;

  pdev->size=*ppos;
  
error:  
  up(&pdev->lock);
  
  return res;
}

ssize_t hello2_read(struct file *pfile, char *buf, size_t count, loff_t *ppos)
{
  HelloDev* pdev=pfile->private_data;
  ssize_t res = 0;
  
  if(count == 0) return 0;
  
  if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;

  if(pdev->data == 0) goto error;
  if(*ppos >= pdev->size) goto error;
  if((*ppos+count) > pdev->size) count=pdev->size-*ppos;
    
  if(copy_to_user(buf, (char*)(pdev->data)+*ppos, count))
  {
    res=-EFAULT;
    goto error;
  }
  *ppos+=count;
  res=count;
  
error:  
  up(&pdev->lock);
  
  return res;
}

void hello_vm_open(struct vm_area_struct * area)
{
  printk(DEVICE_NAME " vm_open.\n");
}

void hello_vm_close(struct vm_area_struct * area)
{
  printk(DEVICE_NAME " vm_close.\n");
}

static struct vm_operations_struct hello_vm_ops =
{
  open: hello_vm_open,
  close: hello_vm_close
};

static int hello_mmap(struct file* pfile, struct vm_area_struct *vma)
{
  HelloDev* pdev=pfile->private_data;
  unsigned long vsize = vma->vm_end - vma->vm_start;
  
  if(vsize != PAGE_SIZE)
  {
    printk(DEVICE_NAME " invalid size: %lu\n", vsize);
    return -EINVAL;
  }
  
  if (vma->vm_flags & VM_WRITE)
    return -EPERM;
  
  if (PAGE_SIZE > (1 << 16))
    return -ENOSYS;
  
  if(remap_pfn_range(vma, vma->vm_start, 
    virt_to_phys((void *)pdev->data) >> PAGE_SHIFT,
		     vma->vm_end - vma->vm_start, vma->vm_page_prot))
  {
    return -EAGAIN;
  }
  vma->vm_ops = &hello_vm_ops;
  hello_vm_open(vma);
  return 0;
}

ssize_t hello3_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos)
{
  HelloDev* pdev=pfile->private_data;
  ssize_t res = -ENOMEM;
  
  if(count == 0) return 0;
  if(*ppos >= BUFFER_SIZE) return -ENOMEM;
  
  if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;

  if(pdev->data == 0) goto error;
  
  if(*ppos == 0) *ppos=pdev->size;
  if((*ppos+count) > BUFFER_SIZE) count=BUFFER_SIZE-*ppos;
    
  if(copy_from_user((char*)(pdev->data)+*ppos, buf, count))
  {
    res=-EFAULT;
    goto error;
  }
  *ppos+=count;
  res=count;

  pdev->size=*ppos;
  
error:  
  up(&pdev->lock);

  if(res>0) wake_up_interruptible(&hello_queue);
  
  return res;
}

ssize_t hello3_read(struct file *pfile, char *buf, size_t count, loff_t *ppos)
{
  HelloDev* pdev=pfile->private_data;
  ssize_t res = 0;
  
  if(count == 0) return 0;
  
  if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;

  if(pdev->data == 0) goto error;
  if((count + *ppos) > BUFFER_SIZE) count = BUFFER_SIZE-*ppos;
  
  if((count + *ppos) > pdev->size)
  {
    up(&pdev->lock);
    
    if (pfile->f_flags & O_NONBLOCK)
      return -EAGAIN;
      
    if(wait_event_interruptible(hello_queue, ((count + *ppos) <= pdev->size))) return -ERESTARTSYS;
    
    if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;
  }
    
  if(copy_to_user(buf, (char*)(pdev->data)+*ppos, count))
  {
    res=-EFAULT;
    goto error;
  }
  *ppos+=count;
  res=count;
  
error:  
  up(&pdev->lock);
  
  return res;
}

unsigned int hello3_poll(struct file *pfile, poll_table *wait)
{
  HelloDev* pdev=pfile->private_data;
  unsigned int mask = 0;

  poll_wait(pfile, &hello_queue,  wait);
  if (pdev->size > 0) mask |= POLLIN | POLLRDNORM;

  return mask;
}


long hello_ioctl(struct file *pfile, 
	unsigned int cmd, unsigned long arg)
{
  HelloDev* pdev=pfile->private_data;
  int res=0;
  
  if(_IOC_TYPE(cmd) != HELLO_IOC_MAGIC) return -ENOTTY;
  if(_IOC_DIR(cmd) & _IOC_READ)
  {
    if(!access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd))) return -EFAULT;
  }
  else if(_IOC_DIR(cmd) & _IOC_WRITE)
  {
    if(!access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd)))  return -EFAULT;
  }
 
  switch(cmd)
  {
    case HELLO_IOC_CLEAR:
      if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;
      pdev->size=0;
      up(&pdev->lock);
      break;
    case HELLO_IOC_GSIZE:
      if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;
      res=put_user(pdev->size, (unsigned long*)arg);
      up(&pdev->lock);
      break;
    case HELLO_IOC_SSIZE:
      if(down_interruptible(&pdev->lock)) return -ERESTARTSYS;
      res=get_user(pdev->size, (unsigned long*)arg);
      up(&pdev->lock);
      break;
    default:
      return -ENOIOCTLCMD;
  }
  return res;
}

loff_t hello_llseek(struct file *pfile, loff_t off, int whence)
{
  HelloDev* pdev=pfile->private_data;
  loff_t res=0;

  switch(whence)
  {
    case 0: /* SEEK_SET */
      res = off;
      break;

    case 1: /* SEEK_CUR */
      res = pfile->f_pos + off;
      break;

    case 2: /* SEEK_END */
      res = pdev->size + off;
      break;

    default: /* can't happen */
      return -EINVAL;
  }
  if (res<0) return -EINVAL;
  pfile->f_pos = res;
  return res;
}

/* A __exit nem használható, mert a hello_init() függvényben hivatkozunk rá. */
void hello_exit(void)
{
  int i;

  device_destroy(hello_class, MKDEV(major_num, 0));
  device_destroy(hello_class, MKDEV(major_num, 1));
  device_destroy(hello_class, MKDEV(major_num, 2));
  class_unregister(hello_class);
  class_destroy(hello_class);
  
  for(i=0; i<NUM_OF_DEVS; i++) 
  {
    //if(helloDevs[i].data!=NULL) kfree(helloDevs[i].data);
    if(helloDevs[i].data!=0) free_page(helloDevs[i].data);
    helloDevs[i].data=0;
    helloDevs[i].size=0;
  }

  unregister_chrdev(major_num, DEVICE_NAME);
  printk(KERN_INFO DEVICE_NAME " felszabaditva.\n");
}

int __init hello_init(void)
{
  int res;
  int i;
  struct device* err;

  res = register_chrdev(major_num, DEVICE_NAME, &hello1_fops);
  if(res<0)
  {
    printk(KERN_WARNING DEVICE_NAME " a major szam: %d nem elerheto\n",major_num);
    return res;
  }
  if(major_num==0) major_num=res;

  hello_class = class_create(THIS_MODULE, CLASS_NAME);
  err = device_create(hello_class, NULL, MKDEV(major_num, 0), NULL, DEVICE_NAME"0");
  err = device_create(hello_class, NULL, MKDEV(major_num, 1), NULL, DEVICE_NAME"1");
  err = device_create(hello_class, NULL, MKDEV(major_num, 2), NULL, DEVICE_NAME"2");
  
  init_waitqueue_head(&hello_queue);
  
  for(i=0; i<NUM_OF_DEVS; i++) 
  {
    helloDevs[i].data=0;
    helloDevs[i].size=0;
    sema_init(&helloDevs[i].lock, 1);
  }
  for(i=0; i<NUM_OF_DEVS; i++) 
  {
    //helloDevs[i].data=kzalloc(BUFFER_SIZE, GFP_KERNEL);
    helloDevs[i].data=get_zeroed_page(GFP_KERNEL);
    if(helloDevs[i].data==0)
    {
      res=-ENOMEM;
      goto error;
    }
  }

  printk(KERN_INFO DEVICE_NAME " betoltve.\n");
  return 0;
  
error:
  hello_exit();
  return res;
}

module_init(hello_init);
module_exit(hello_exit);

MODULE_DESCRIPTION("Adv Hello driver");
MODULE_LICENSE("GPL");
