/* hellodriver.c
 *
 * Egyszerű karakteres eszközkezelő modul.
 * 
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/device.h>

#define DEVICE_NAME "hello"
static int major_num=120;
#define CLASS_NAME "helloclass"
static struct class* hello_class;

#define MSG_LEN 27
static char* msg="Hello! How are you doing?\n";

/* Az írás műveletet kezelő rutin. */
static ssize_t hello_write(struct file *pfile, const char *buf, size_t count, loff_t *ppos)
{
  /* A kernel memoriából lefoglaljuk a szükséges területet. */
  char* str=kmalloc(count+1,GFP_KERNEL);

  /* Átmásoljuk a felhasználói címterületről az adatokat a kernel címterületére. */
  if(!copy_from_user(str,buf,count))
  {
    /* Lezárjuk egy 0 karakterrel. */
    str[count]=0;

    /* A printk() függvényhívással kiírjuk. */
    printk(DEVICE_NAME " write: %s\n",str);
  }
  
  /* Felszabadítjuk a lefoglalt területet. */
  kfree(str);

  /* Visszatérünk az olvasott byte-ok számával. */
  return count;
}

/* Az olvasás műveletet kezelő függvény. */
static ssize_t hello_read(struct file *pfile, char *buf, size_t count, loff_t *ppos)
{
  /* Ha a pozíció a szöveg végére mutat, akkor visszatérünk 0 értékkel (EOF). */
  if(*ppos >= MSG_LEN) return 0;

  /* Maximum annyi byte-ot szolgálunk ki, amilyen hosszú a szöveg. */
  if((*ppos+count) > MSG_LEN) count=MSG_LEN-*ppos;
  
  /* Átmásoljuk az adatot a kernel címterületről, az olvasó függvény által kapott bufferbe. */
  if(copy_to_user(buf, msg+*ppos, count))
  {
    return -EFAULT;
  }
  *ppos+=count;
  
  /* Visszatérünk a visszaadott adatmennyiség hosszával. */
  return count;
}

/* A megnyitás műveletet kezelő függvény. */
static int hello_open(struct inode *inode, struct file *pfile)
{
  /* Növeljuk a használati számlálót. */
  try_module_get(THIS_MODULE);
  printk(DEVICE_NAME " open.\n");
  return 0;
}

/* A lezárás műveletet kezelő függvény. */
static int hello_close(struct inode *inode, struct file *pfile)
{
  printk(DEVICE_NAME " close.\n");
  /* Csökkentjük a használati számlálót. */
  module_put(THIS_MODULE);
  return 0;
}

static struct file_operations hello_fops =
{
  owner:	THIS_MODULE,
  read:	hello_read,
  write:	hello_write,
  open:	hello_open,
  release:	hello_close
};

/* A kernel modul inicializálását végző függvény. */
static int __init hello_init(void)
{
  int res;
  struct device* err;
  
  /* Regisztráljuk a karakter típusú eszközt. */
  res = register_chrdev(major_num, DEVICE_NAME, &hello_fops);
  if(res < 0)
  {
    printk(DEVICE_NAME " a major szam: %d nem elerheto\n", major_num);
    return res;
  }
  if(major_num == 0) major_num = res;

  /* Az udev számára jelzés, hogy hozza létre az eszközállományt. */
  hello_class = class_create(THIS_MODULE, CLASS_NAME);
  err = device_create(hello_class, NULL, MKDEV(major_num, 0), NULL, DEVICE_NAME);
  
  printk(DEVICE_NAME " betoltve. Major szam: %d\n", major_num);
  return 0;
}

/* A kernel modul eltávolítása előtt a felszabadításokat végzi. */
static void __exit hello_exit(void)
{
  /* Az udev számára jelzés, az eszköz eltávolításáról. */
  device_destroy(hello_class, MKDEV(major_num, 0));
  class_unregister(hello_class);
  class_destroy(hello_class);
  
  /* Eltávolítjuk az eszköz regisztrációját. */
  unregister_chrdev(major_num, DEVICE_NAME);
  printk(DEVICE_NAME " felszabaditva.\n");	
}

module_init(hello_init);
module_exit(hello_exit);

MODULE_DESCRIPTION("Hello driver");
MODULE_LICENSE("GPL");
